package dsn

import (
	"errors"
	"fmt"
	"strconv"
	"strings"

	"gitlab.wise-paas.com/sdk/configcenter/util"
)

var (
	rabbitmqSchema     = "rabbitmq"
	ErrInvalidRabbitmq = errors.New("invalid rabbitmq")
	ErrInvalidPort     = errors.New("invalid port")
)

type ConnectInfo struct {
	Schema   string
	Host     string
	Port     int
	Username string
	Password string
	Database string
	Params   map[string]string
	Raw      string
}

func (c *ConnectInfo) ToMqtt() error {
	if strings.ToLower(c.Schema) != rabbitmqSchema {
		return ErrInvalidRabbitmq
	}

	mqttPortStr := c.Params["mqtt"]
	mqttPort, err := strconv.Atoi(mqttPortStr)
	if err != nil {
		return ErrInvalidPort
	}

	c.Username = c.Database + ":" + c.Username
	c.Port = mqttPort
	return nil
}

func (c *ConnectInfo) ToAMQP() error {
	if strings.ToLower(c.Schema) != rabbitmqSchema {
		return ErrInvalidRabbitmq
	}

	amqpPortStr := c.Params["amqp"]
	amqpPort, err := strconv.Atoi(amqpPortStr)
	if err != nil {
		return ErrInvalidPort
	}

	c.Port = amqpPort
	return nil
}

func (c *ConnectInfo) ToMqtts() error {
	if strings.ToLower(c.Schema) != rabbitmqSchema {
		return ErrInvalidRabbitmq
	}

	mqttPortStr := c.Params["mqtts"]
	mqttPort, err := strconv.Atoi(mqttPortStr)
	if err != nil {
		return ErrInvalidPort
	}

	c.Username = c.Database + ":" + c.Username
	c.Port = mqttPort
	return nil
}

func (c *ConnectInfo) ToAMQPs() error {
	if strings.ToLower(c.Schema) != rabbitmqSchema {
		return ErrInvalidRabbitmq
	}

	amqpPortStr := c.Params["amqps"]
	amqpPort, err := strconv.Atoi(amqpPortStr)
	if err != nil {
		return ErrInvalidPort
	}

	c.Port = amqpPort
	return nil
}

func (c *ConnectInfo) ToMqttClone() (ConnectInfo, error) {
	newConnInfo := ConnectInfo{}
	util.CopyToTarget(c, &newConnInfo)

	if err := newConnInfo.ToMqtt(); err != nil {
		return ConnectInfo{}, err
	}

	return newConnInfo, nil
}

func (c *ConnectInfo) ToMqttsClone() (ConnectInfo, error) {
	newConnInfo := ConnectInfo{}
	util.CopyToTarget(c, &newConnInfo)

	if err := newConnInfo.ToMqtts(); err != nil {
		return ConnectInfo{}, err
	}

	return newConnInfo, nil
}

func (c *ConnectInfo) ToAMQPClone() (ConnectInfo, error) {
	newConnInfo := ConnectInfo{}
	util.CopyToTarget(c, &newConnInfo)

	if err := newConnInfo.ToAMQP(); err != nil {
		return ConnectInfo{}, err
	}

	return newConnInfo, nil
}

func (c *ConnectInfo) ToAMQPsClone() (ConnectInfo, error) {
	newConnInfo := ConnectInfo{}
	util.CopyToTarget(c, &newConnInfo)

	if err := newConnInfo.ToAMQPs(); err != nil {
		return ConnectInfo{}, err
	}

	return newConnInfo, nil
}

func ParseDSN(dsn string) (ConnectInfo, error) {
	ret := ConnectInfo{
		Raw: dsn,
	}

	paramIndex := strings.LastIndex(dsn, "?")
	if paramIndex >= 0 {
		ret.Params = map[string]string{}
		params := dsn[paramIndex+1:]
		length := len(params)
		index := 0
		wantKey := true
		wantValue := false
		key := ""
		value := ""

		for i := 0; i < length; i++ {
			switch params[i] {
			case '=':
				if !wantKey {
					return ret, fmt.Errorf("invalid params")
				}

				key = params[index:i]
				i++
				index = i
				wantKey = false
				wantValue = true
			case '&':
				if !wantValue {
					return ret, fmt.Errorf("invalid params")
				}

				value = params[index:i]
				i++
				index = i
				wantKey = true
				wantValue = false

				if key == "" {
					return ret, fmt.Errorf("invalid params")
				}

				ret.Params[key] = value
			default:
			}
		}

		if wantValue {
			if key == "" {
				return ret, fmt.Errorf("invalid params")
			}
			value = params[index:]
			ret.Params[key] = value
		}

		dsn = dsn[:paramIndex]
	}

	index := strings.Index(dsn, "://")
	if index < 0 {
		return ret, fmt.Errorf("lack of schema")
	}

	ret.Schema = dsn[:index]
	dsn = dsn[index+3:]

	index = strings.LastIndex(dsn, "@")
	if index > 0 {
		usernamePassword := dsn[:index]
		dsn = dsn[index+1:]
		index = strings.Index(usernamePassword, ":")
		if index < 0 {
			return ret, fmt.Errorf("username or password format error")
		}

		ret.Username = usernamePassword[:index]
		ret.Password = usernamePassword[index+1:]
	}

	index = strings.Index(dsn, "/")
	if index > 0 {
		ret.Database = dsn[index+1:]
		dsn = dsn[:index]
	}

	index = strings.Index(dsn, ":")
	if index > 0 {
		ret.Host = dsn[:index]
		port, err := strconv.Atoi(dsn[index+1:])
		if err != nil {
			return ret, err
		}

		ret.Port = port
	} else {
		ret.Host = dsn
	}

	return ret, nil
}
