package dsn

import (
	"reflect"
	"testing"

	"github.com/stretchr/testify/assert"
)

var (
	testCase []ConnectInfo = []ConnectInfo{
		{
			Schema:   "postgres",
			Username: "admin",
			Password: "admin",
			Host:     "127.0.0.1",
			Port:     5432,
			Database: "postgres",
			Raw:      "postgres://admin:admin@127.0.0.1:5432/postgres",
		},
		{
			Schema:   "postgres",
			Host:     "127.0.0.1",
			Port:     5432,
			Database: "postgres",
			Raw:      "postgres://127.0.0.1:5432/postgres",
		},
		{
			Schema: "sqlite",
			Host:   "127.0.0.1",
			Raw:    "sqlite://127.0.0.1",
		},
		{
			Schema:   "postgres",
			Host:     "127.0.0.1",
			Port:     5432,
			Username: "admin",
			Password: "admin",
			Params:   map[string]string{"a": "b1", "c": "d1"},
			Raw:      "postgres://admin:admin@127.0.0.1:5432?a=b1&c=d1",
		},
		{
			Schema:   "postgres",
			Host:     "127.0.0.1",
			Port:     5432,
			Database: "pg",
			Username: "admin",
			Password: "admin",
			Params:   map[string]string{"a": "b1", "c": "d1"},
			Raw:      "postgres://admin:admin@127.0.0.1:5432/pg?a=b1&c=d1",
		},
		{
			Schema:   "schema",
			Host:     "127.0.0.1",
			Port:     8080,
			Password: "pwd",
			Raw:      "schema://:pwd@127.0.0.1:8080",
		},
	}
)

func TestParseDSN(t *testing.T) {
	for _, connectionInfo := range testCase {
		parsedConnectionInfo, err := ParseDSN(connectionInfo.Raw)
		if err != nil {
			t.Error(err)
			return
		}

		if !reflect.DeepEqual(parsedConnectionInfo, connectionInfo) {
			t.Log(parsedConnectionInfo)
			t.Log(connectionInfo)

			t.Errorf("test failed for dsn '%s'", connectionInfo.Raw)
			return
		}
	}
}

func TestToMQTT(t *testing.T) {
	connInfo := ConnectInfo{
		Schema:   rabbitmqSchema,
		Host:     "127.0.0.1",
		Port:     5432,
		Username: "admin",
		Password: "admin",
		Database: "db",
		Params:   map[string]string{"mqtt": "1883", "c": "d1"},
	}

	if err := connInfo.ToMqtt(); err != nil {
		t.Error(err)
		return
	}

	assert.Equal(t, connInfo.Port, 1883)
	assert.Equal(t, connInfo.Username, "db:admin")
}

func TestToAMQP(t *testing.T) {
	connInfo := ConnectInfo{
		Schema:   rabbitmqSchema,
		Host:     "127.0.0.1",
		Port:     5432,
		Username: "admin",
		Password: "admin",
		Database: "db",
		Params:   map[string]string{"amqp": "5672", "c": "d1"},
	}

	if err := connInfo.ToAMQP(); err != nil {
		t.Error(err)
		return
	}

	assert.Equal(t, connInfo.Port, 5672)
}

func TestToMQTTClone(t *testing.T) {
	connInfo := ConnectInfo{
		Schema:   rabbitmqSchema,
		Host:     "127.0.0.1",
		Port:     5432,
		Username: "admin",
		Password: "admin",
		Database: "db",
		Params:   map[string]string{"mqtt": "1883", "c": "d1"},
	}

	newConnInfo, err := connInfo.ToMqttClone()
	if err != nil {
		t.Error(err)
		return
	}

	assert.Equal(t, connInfo.Port, 5432)
	assert.Equal(t, connInfo.Username, "admin")

	assert.Equal(t, newConnInfo.Port, 1883)
	assert.Equal(t, newConnInfo.Username, "db:admin")
}

func TestToAMQPClone(t *testing.T) {
	connInfo := ConnectInfo{
		Schema:   rabbitmqSchema,
		Host:     "127.0.0.1",
		Port:     5432,
		Username: "admin",
		Password: "admin",
		Database: "db",
		Params:   map[string]string{"amqp": "5672", "c": "d1"},
	}

	newConnInfo, err := connInfo.ToAMQPClone()
	if err != nil {
		t.Error(err)
		return
	}

	assert.Equal(t, connInfo.Port, 5432)
	assert.Equal(t, newConnInfo.Port, 5672)
}
