package module

import (
	"context"
	"errors"
	"fmt"
	"os"
	"strings"
	"sync"
	"time"

	"github.com/wh8199/log"

	"gitlab.wise-paas.com/sdk/configcenter/pkg/configcenter"
	"gitlab.wise-paas.com/sdk/configcenter/portrange"
	"google.golang.org/grpc"
)

type ModuleConfig struct {
	Name string
	// default value is "10086-10090"
	PortRange string
	// all service instances needed by this module
	ServiceInstances []*configcenter.ServiceInstance
	// all service routers
	Routers []*configcenter.Router
}

type ModuleStatus int

func (m ModuleStatus) String() string {
	switch m {
	case Connecting:
		return "Connecting"
	case DisConnected:
		return "DisConnected"
	case Reconnecting:
		return "Reconnecting"
	case Connected:
		return "Connected"
	case Destry:
		return "Destry"
	default:
		return "Unknown"
	}
}

const (
	Connecting = iota
	DisConnected
	Reconnecting
	Connected
	Destry
)

type Module struct {
	moduleConfig ModuleConfig
	ctx          context.Context

	lock sync.RWMutex
	// configs are read from config center
	configs map[string]string
	// services are read from config center
	services map[string]*configcenter.ServiceInfo
	// service instance
	serviceInstances map[string]string
	status           ModuleStatus
	addr             string
}

func NewModule(moduleConfig ModuleConfig, ctx context.Context) Module {
	if moduleConfig.Name == "" {
		log.Fatal("empty module id")
	}

	return Module{
		moduleConfig: moduleConfig,
		ctx:          ctx,
	}
}

func (m *Module) findRuntime() string {
	if os.Getenv("CONFIG_CENTER_ADDRESS") != "" {
		return os.Getenv("CONFIG_CENTER_ADDRESS")
	}

	portRange := m.moduleConfig.PortRange
	if portRange == "" {
		portRange = "10086-10090"
	}

	min, max, err := portrange.ParsePortRange(portRange)
	if err != nil {
		log.Error(err)
		return ""
	}

	ret := FindCandidate(func(host string, port uint16) bool {
		ctx, cancelFunc := context.WithTimeout(context.Background(), 100*time.Millisecond)
		defer cancelFunc()

		conn, client, err := dial(fmt.Sprintf("%s:%d", host, port), ctx)
		if err != nil {
			return false
		}

		defer conn.Close()

		_, err = client.GetPublicConfigs(ctx, &configcenter.QueryConfigRequest{})
		return err == nil
	}, min, max)

	if len(ret) == 0 {
		return ""
	}

	addr := fmt.Sprintf("127.0.0.1:%d", ret[0])
	m.addr = addr

	return addr
}

func dial(addr string, ctx context.Context) (*grpc.ClientConn, configcenter.ConfigCenterClient, error) {
	if addr == "" {
		return nil, nil, fmt.Errorf("empty address")
	}

	conn, err := grpc.DialContext(ctx, addr, grpc.WithInsecure(), grpc.FailOnNonTempDialError(true))
	if err != nil {
		log.Info(err)
		return nil, nil, err
	}

	client := configcenter.NewConfigCenterClient(conn)

	return conn, client, nil
}

func (m *Module) linkConfigCenterOnce(addr string) error {
	conn, client, err := dial(addr, m.ctx)
	if err != nil {
		if err.Error() == context.Canceled.Error() {
			return context.Canceled
		}
		return err
	}

	defer conn.Close()

	m.status = Connected

	for {
		ticker := time.NewTicker(time.Second * 1)
		defer ticker.Stop()

		for {
			select {
			case <-m.ctx.Done():
				return context.Canceled
			case <-ticker.C:
				cfgs, err := client.GetPublicConfigs(m.ctx, &configcenter.QueryConfigRequest{})
				if err != nil {
					// this error is a network error, so return and reconnect
					log.Error(err)
					return err
				}

				m.lock.Lock()
				m.configs = cfgs.Configs
				m.lock.Unlock()

				services, err := client.GetServices(m.ctx, &configcenter.QueryServiceRequest{})
				if err != nil {
					log.Error(err)
					return err
				}

				m.lock.Lock()
				m.services = services.Infos
				m.lock.Unlock()

				// keep alive
				resp, err := client.ServiceRegistration(m.ctx, &configcenter.ServiceRegistrationRequest{
					ServiceName:      m.moduleConfig.Name,
					ServiceInstances: m.moduleConfig.ServiceInstances,
					Routers:          m.moduleConfig.Routers,
				})
				if err != nil {
					// this error maybe service registration failed, not a network error
					log.Error(err)
					continue
				}

				m.lock.Lock()
				m.serviceInstances = resp.ServiceInstances
				m.lock.Unlock()
			}
		}
	}
}

func (m *Module) LinkConfigCenter() error {
	addr := m.findRuntime()

	m.status = Connecting

	for {
		err := m.linkConfigCenterOnce(addr)
		if errors.Is(err, context.Canceled) {
			m.status = Destry
			return nil
		}

		m.status = Reconnecting

		addr = m.findRuntime()
		log.Error(err)
		time.Sleep(time.Second * 5)
	}
}

func (m *Module) GetPublicConfigsUnsafe() map[string]string {
	m.lock.RLock()
	defer m.lock.RUnlock()

	return m.configs
}

func (m *Module) GetPublicConfig(key string) (string, bool) {
	m.lock.RLock()
	defer m.lock.RUnlock()

	value, ok := m.configs[key]
	return value, ok
}

func (m *Module) GetServiceInstancesUnsafe() map[string]string {
	m.lock.RLock()
	defer m.lock.RUnlock()

	return m.serviceInstances
}

func (m *Module) GetServiceInstance(key string) (string, bool) {
	m.lock.RLock()
	defer m.lock.RUnlock()

	value, ok := m.serviceInstances[strings.ToLower(key)]
	return value, ok
}

func (m *Module) GetAllServicesUnsafe() map[string]*configcenter.ServiceInfo {
	m.lock.RLock()
	defer m.lock.RUnlock()

	return m.services
}

func (m *Module) GetRouters(serviceName string) (map[string]*configcenter.Router, bool) {
	m.lock.RLock()
	defer m.lock.RUnlock()

	service, ok := m.services[serviceName]
	if !ok {
		return nil, false
	}

	return service.Routers, true
}

func (m *Module) GetRouter(serviceName string, routerName string) (*configcenter.Router, bool) {
	m.lock.RLock()
	defer m.lock.RUnlock()

	service, ok := m.services[serviceName]
	if !ok {
		return nil, false
	}

	if service.Routers == nil {
		return nil, false
	}

	router, ok := service.Routers[routerName]
	if !ok {
		return nil, false
	}

	return router, true
}

func (m *Module) GetRouterPublicAddress(serviceName string, routerName string) (string, bool) {
	router, ok := m.GetRouter(serviceName, routerName)
	if !ok {
		return "", false
	}

	return router.PublicAddress, true
}

func (m *Module) GetStatus() ModuleStatus {
	return m.status
}

func (m *Module) UpdateConfig(cfg ModuleConfig) {
	m.lock.Lock()
	defer m.lock.Unlock()

	m.moduleConfig = cfg
}

func (m *Module) GetConfigCenterAddress() string {
	if m.addr == "" {
		return os.Getenv("CONFIG_CENTER_ADDRESS")
	}

	return m.addr
}

func (m *Module) GetConfigCenterAddressRestful() string {
	if m.addr == "" {
		return "http://" + os.Getenv("CONFIG_CENTER_ADDRESS")
	}

	return "http://" + m.addr
}
